/*
 * Copyright (C) 2009 Tudor Marian (tudorm@cs.cornell.edu)
 * 
 */

#include "s3fs.h"
#include "s3ops.h"

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/time.h>
#include <time.h>

//#define NDEBUG /* disables assert if macro defined above #include */
#include <assert.h>

struct s3fs_buckets *sb;

struct dirbuf
{
	char *p;
	size_t size;
};

/* inode number 1 belongs to /, fuse assumes this */

/* format filesystem --- superblock, and inode 1 (or /) */
int s3fs_format(char *inode_bucket, char *data_bucket)
{
	return 1;
}

/* 
 * initialize filesystem --- e.g. read the superblock
 * @param userdata the user data passed to fuse_lowlevel_new()
 */
void s3fs_init(void *userdata, struct fuse_conn_info *conn)
{
	sb = (struct s3fs_buckets *) userdata;
	/* do other stuff */
}

/* To give you a head start, what follows is the typical fuse example  */
static const char *hello_str = "Hello World!\n";
static const char *hello_name = "hello";

static int do_stat(fuse_ino_t ino, struct stat *stbuf)
{
	stbuf->st_ino = ino;
	switch (ino)
	{
		case 1:
			stbuf->st_mode = S_IFDIR | 0755;
			stbuf->st_nlink = 2;
			break;

		case 2:
			stbuf->st_mode = S_IFREG | 0444;
			stbuf->st_nlink = 1;
			stbuf->st_size = strlen(hello_str);
			break;

		default:
			return -1;
	}
	return 0;
}

void s3fs_getattr(fuse_req_t req, fuse_ino_t ino,
		struct fuse_file_info *fi)
{
	struct stat stbuf;
	memset(&stbuf, 0, sizeof(stbuf));

	if (do_stat(ino, &stbuf) == -1)
		fuse_reply_err(req, ENOENT);
	else
		fuse_reply_attr(req, &stbuf, 1.0);
}

void s3fs_lookup(fuse_req_t req, fuse_ino_t parent, const char *name)
{
	struct fuse_entry_param e;

	if (parent != 1 || strcmp(name, hello_name) != 0)
		fuse_reply_err(req, ENOENT);
	else
	{
		memset(&e, 0, sizeof(e));
		e.ino = 2;
		e.attr_timeout = 1.0;
		e.entry_timeout = 1.0;
		do_stat(e.ino, &e.attr);
		fuse_reply_entry(req, &e);
	}
}

/* Miklos's original dirbuf_add code, I am not a big fan of realloc */
static void dirbuf_add(fuse_req_t req, struct dirbuf *b, const char *name,
		fuse_ino_t ino)
{
	struct stat stbuf;
	size_t oldsize = b->size;
	b->size += fuse_add_direntry(req, NULL, 0, name, NULL, 0);
	char *newp = (char *)realloc(b->p, b->size);
	if (!newp)
	{
		fprintf(stderr, "*** fatal error: cannot allocate memory\n");
		abort();
	}
	b->p = newp;
	memset(&stbuf, 0, sizeof(stbuf));
	stbuf.st_ino = ino;
	fuse_add_direntry(req, b->p + oldsize, b->size - oldsize, name, &stbuf,
			b->size);
}

#define min(x, y) ((x) < (y) ? (x) : (y))

static int reply_buf_limited(fuse_req_t req, const char *buf, size_t bufsize,
		off_t off, size_t maxsize)
{
	if (off < bufsize)
		return fuse_reply_buf(req, buf + off, min(bufsize - off, maxsize));
	else
		return fuse_reply_buf(req, NULL, 0);
}

void s3fs_readdir(fuse_req_t req, fuse_ino_t ino, size_t size,
		off_t off, struct fuse_file_info *fi)
{
	if (ino != 1)
		fuse_reply_err(req, ENOTDIR);
	else
	{
		struct dirbuf b;

		memset(&b, 0, sizeof(b));
		dirbuf_add(req, &b, ".", 1);
		dirbuf_add(req, &b, "..", 1);
		dirbuf_add(req, &b, hello_name, 2);
		reply_buf_limited(req, b.p, b.size, off, size);
		free(b.p);
	}
}

void s3fs_open(fuse_req_t req, fuse_ino_t ino,
		struct fuse_file_info *fi)
{
	if (ino != 2)
		fuse_reply_err(req, EISDIR);
	else if ((fi->flags & 3) != O_RDONLY)
		fuse_reply_err(req, EACCES);
	else
		fuse_reply_open(req, fi);
}

void s3fs_read(fuse_req_t req, fuse_ino_t ino, size_t size,
		off_t off, struct fuse_file_info *fi)
{
	assert(ino == 2);
	reply_buf_limited(req, hello_str, strlen(hello_str), off, size);
}


